

from cefkivy.handlers.base import ClientHandlerBase
from ..components import jsdialog

from cefkivy.browser import cefpython


class JavascriptDialogHandler(ClientHandlerBase):
    # https://github.com/cztomczak/cefpython/blob/master/api/JavascriptDialogHandler.md
    def OnJavascriptDialog(self, browser, origin_url, dialog_type,
                           message_text, default_prompt_text, callback,
                           suppress_message_out):
        kwargs = {
            'browser': browser,
            'origin_url': origin_url,
            'message_text': message_text,
            'callback': callback
        }
        dialog = None
        if dialog_type == cefpython.JSDIALOGTYPE_ALERT:
            dialog = jsdialog.JSDialogAlert(**kwargs)
        elif dialog_type == cefpython.JSDIALOGTYPE_CONFIRM:
            dialog = jsdialog.JSDialogConfirm(**kwargs)
        elif dialog_type == cefpython.JSDIALOGTYPE_PROMPT:
            kwargs['default_prompt_text'] = default_prompt_text
            dialog = jsdialog.JSDialogPrompt(**kwargs)
        if dialog:
            self._widget.dialog_show(dialog)
        suppress_message_out.append(True)
        return True

    def OnBeforeUnloadJavascriptDialog(self, browser, message_text, is_reload, callback):
        return False

    def OnResetJavascriptDialogState(self, browser):
        pass

    def OnJavascriptDialogClosed(self, browser):
        pass



from cefpython3 import cefpython
from kivy import Logger

from cefkivy.handlers.base import ClientHandlerBase


class RequestHandler(ClientHandlerBase):
    def GetAuthCredentials(self, *args, **kwargs):
        pass

    def GetCookieManager(self, browser, main_url):
        cookie_manager = cefpython.CookieManager.GetGlobalManager()
        if cookie_manager:
            return cookie_manager
        else:
            Logger.warn("No cookie manager found!")

    def GetResourceHandler(self, browser, frame, request):
        pass

    def OnBeforeBrowse(self, browser, frame, request, user_gesture, is_redirect):
        pass

    def OnBeforeResourceLoad(self, browser, frame, request):
        pass

    def OnQuotaRequest(self, *args, **kwargs):
        pass

    def OnResourceRedirect(self, *args, **kwargs):
        pass

    def OnProtocolExecution(self, browser, url, allow_execution_out):
        pass

    def OnRendererProcessTerminated(self, browser, status):
        pass



import cefpython3.cefpython_py37

from . import keycodes


class KeystrokeEventProcessor(object):
    def __init__(self, cefpython, browser_widget, *args, **kwargs):
        self.cefpython = cefpython
        self.browser_widget = browser_widget
        self._use_textinput = kwargs.pop('use_textinput', True)
        # Kivy does not provide modifiers in on_key_up, but these
        # must be sent to CEF as well.
        self.is_lshift = False
        self.is_rshift = False
        self.is_lctrl = False
        self.is_rctrl = False
        self.is_lalt = False
        self.is_ralt = False
        self.is_altgr = False

    def reset_all_modifiers(self):
        self.is_lshift = False
        self.is_lctrl = False
        self.is_lalt = False
        self.is_rshift = False
        self.is_rctrl = False
        self.is_ralt = False
        self.is_altgr = False

    def on_key_down(self, browser, keyboard, keycode, text, modifiers):
        # print("\non_key_down:", keycode, text, modifiers)
        if keycodes.get_character_native(keycode[0]) == 'escape':
            # On escape release the keyboard
            self.browser_widget.release_keyboard()
            return

        cef_modifiers = self.cefpython.EVENTFLAG_NONE
        if "shift" in modifiers:
            cef_modifiers |= self.cefpython.EVENTFLAG_SHIFT_DOWN
        if "ctrl" in modifiers:
            cef_modifiers |= self.cefpython.EVENTFLAG_CONTROL_DOWN
        if "alt" in modifiers:
            cef_modifiers |= self.cefpython.EVENTFLAG_ALT_DOWN
        if "capslock" in modifiers:
            cef_modifiers |= self.cefpython.EVENTFLAG_CAPS_LOCK_ON

        try:
            native_key_code = keycodes.get_unix_from_native(keycode[0])
            windows_key_code = keycodes.get_windows_from_native(keycode[0])
            character = keycodes.get_character_native(keycode[0],
                                                      modifiers=modifiers)
            unmodified_character = keycodes.get_character_native(keycode[0])
        except KeyError:
            native_key_code = keycode[0]
            windows_key_code = keycode[0]
            character = keycode[1]
            unmodified_character = keycode[1]

        # Only send KEYEVENT_KEYDOWN if it is a special key (tab, return ...)
        # Eh? Convert every other key to it's utf8 int value and send this as the key
        if keycodes.check_is_special(native_key_code):
            event_type = self.cefpython.KEYEVENT_KEYDOWN
            character = ''
            unmodified_character = ''
        else:
            # # cef_key_code = ord(text)
            # # We have to convert the apostrophes as the utf8 key-code somehow don't get recognized by cef
            # if cef_key_code == 96:
            #     cef_key_code = 39
            # if cef_key_code == 8220:
            #     cef_key_code = 34
            event_type = self.cefpython.KEYEVENT_CHAR

        # When the key is the return key, send it as a KEYEVENT_CHAR as it will not work in textinputs
        # TODO This thing doesnt work
        # if keycodes.get_character_native(native_key_code) == "enter":
        #     event_type = self.cefpython.KEYEVENT_CHAR

        key_event = {
            "type": event_type,
            "native_key_code": native_key_code,
            "modifiers": cef_modifiers,
            "is_system_key": False,
            "windows_key_code": windows_key_code,
        }

        if character and len(character) == 1:
            key_event.update({
                "character": ord(character),
                "unmodified_character": ord(unmodified_character),
            })

        if self._use_textinput and key_event['type'] == self.cefpython.KEYEVENT_CHAR:
            pass
        else:
            # print("keydown keyEvent: %s" % key_event)
            browser.SendKeyEvent(key_event)

        if keycodes.get_character_native(native_key_code) == 'lshift':
            self.is_lshift = True
        if keycodes.get_character_native(native_key_code) == 'rshift':
            self.is_rshift = True
        if keycodes.get_character_native(native_key_code) == 'lctrl':
            self.is_lctrl = True
        if keycodes.get_character_native(native_key_code) == 'rctrl':
            self.is_rctrl = True
        if keycodes.get_character_native(native_key_code) == 'lalt':
            self.is_lalt = True
        if keycodes.get_character_native(native_key_code) == 'ralt':
            self.is_ralt = True
        if keycodes.get_character_native(native_key_code) == 'alt-gr':
            self.is_altgr = True

    def on_key_up(self, browser, keyboard, keycode):
        # if self._use_textinput:
        #     return
        # print("keyup keyEvent %s" % (keycode,))
        cef_modifiers = self.cefpython.EVENTFLAG_NONE
        if self.is_lshift or self.is_rshift:
            cef_modifiers |= self.cefpython.EVENTFLAG_SHIFT_DOWN
        if self.is_lctrl or self.is_rctrl:
            cef_modifiers |= self.cefpython.EVENTFLAG_CONTROL_DOWN
        if self.is_lalt or self.is_ralt or self.is_altgr:
            cef_modifiers |= self.cefpython.EVENTFLAG_ALT_DOWN

        try:
            native_key_code = keycodes.get_unix_from_native(keycode[0])
            windows_key_code = keycodes.get_windows_from_native(keycode[0])
        except KeyError:
            native_key_code = keycode[0]
            windows_key_code = keycode[0]

        if keycodes.check_is_special(native_key_code):
            key_event = {
                "type": self.cefpython.KEYEVENT_KEYUP,
                "native_key_code": native_key_code,
                "modifiers": cef_modifiers,
                "is_system_key": False,
                "windows_key_code": windows_key_code,
            }
            browser.SendKeyEvent(key_event)

        if keycodes.get_character_native(native_key_code) == 'lshift':
            self.is_lshift = False
        if keycodes.get_character_native(native_key_code) == 'rshift':
            self.is_rshift = False
        if keycodes.get_character_native(native_key_code) == 'lctrl':
            self.is_lctrl = False
        if keycodes.get_character_native(native_key_code) == 'rctrl':
            self.is_rctrl = False
        if keycodes.get_character_native(native_key_code) == 'lalt':
            self.is_lalt = False
        if keycodes.get_character_native(native_key_code) == 'ralt':
            self.is_ralt = False
        if keycodes.get_character_native(native_key_code) == 'alt-gr':
            self.is_altgr = False

    def on_textinput(self, browser, keyboard, character):
        key_event = {
            "type": cefpython3.cefpython_py37.KEYEVENT_CHAR,
            "character": ord(character),
            # "native_key_code": native_key_code,
            # "modifiers": cef_modifiers,
            # "is_system_key": False,
            # "windows_key_code": windows_key_code,
        }
        browser.SendKeyEvent(key_event)





from kivy.logger import Logger
from kivy.properties import OptionProperty

from ..browser import cefpython
from ..inputs.keyboard import KeystrokeEventProcessor
from ..components.keyboard import KeyboardManager


class KeyboardMixin(object):
    keyboard_mode = OptionProperty("local", options=("global", "local"))

    def __init__(self, **kwargs):
        Logger.debug("cefkivy: Starting the Keystroke Processor")
        self.keystroke_processor = KeystrokeEventProcessor(
            cefpython=cefpython, browser_widget=self
        )
        Logger.debug("cefkivy: Starting the Keyboard Manager")
        self.keyboard_manager = KeyboardManager(self, self.browser, **kwargs)

        def _propagate_mode(_, mode):
            self.keyboard_manager.keyboard_mode = mode
        self.bind(keyboard_mode=_propagate_mode)

    def release_keyboard(self):
        self.keyboard_manager.release_keyboard()

    def request_keyboard(self):
        self.keyboard_manager.request_keyboard()

    def on_key_down(self, *args):
        # print("Kivy Down Event : ", args)
        self.keystroke_processor.on_key_down(self.browser, *args)

    def on_key_up(self, *args):
        # print("Kivy Up Event : ", args)
        self.keystroke_processor.on_key_up(self.browser, *args)

    def on_textinput(self, *args):
        # print("Kivy TextInput Event : ", args)
        self.keystroke_processor.on_textinput(self.browser, *args)





from collections import namedtuple

KeyCode = namedtuple('keycode', 'unix windows character is_special shift capslock')

# Windows codes from https://boostrobotics.eu/windows-key-codes/
# Unix codes from Dell...

# Obviously, gave up entirely on trying to keep windows codes straight when shift
# etc. came into play.

_codes = [
    KeyCode(301, 20, 'capslock', True, 0, False),
    KeyCode(27, 27, 'escape', True, 0, False),
    KeyCode(8, 8, 'backspace', True, 0, False),
    KeyCode(9, 9, 'tab', True, 0, False),
    KeyCode(13, 13, 'enter', True, 0, False),
    KeyCode(271, 13, 'enter', True, 0, False),
    KeyCode(304, 160, 'lshift', True, 0, False),
    KeyCode(303, 161, 'rshift', True, 0, False),
    KeyCode(305, 162, 'lctrl', True, 0, False),
    KeyCode(306, 163, 'rctrl', True, 0, False),
    KeyCode(308, 164, 'lalt', True, 0, False),
    KeyCode(313, 165, 'ralt', True, 0, False),
    KeyCode(307, 18, 'alt-gr', True, 0, False), # NOTE AltGr and ralt are actually different keys.
    KeyCode(277, 45, 'insert', True, 0, False),
    KeyCode(127, 46, 'delete', True, 0, False),
    KeyCode(278, 36, 'home', True, 0, False),
    KeyCode(279, 35, 'end', True, 0, False),
    KeyCode(280, 33, 'pageup', True, 0, False),
    KeyCode(281, 34, 'pagedown', True, 0, False),
    KeyCode(276, 37, 'left', True, 0, False),
    KeyCode(273, 38, 'up', True, 0, False),
    KeyCode(275, 39, 'right', True, 0, False),
    KeyCode(274, 40, 'down', True, 0, False),
    KeyCode(32, 32, ' ', False, 0, False),
    KeyCode(91, 91, '[', False, ord('{') - 91, False),
    KeyCode(93, 93, ']', False, ord('}') - 93, False),
    KeyCode(59, 59, ';', False, ord(':') - 59, False),
    KeyCode(39, 39, "'", False, ord('"') - 39, False),
    KeyCode(47, 47, '/', False, ord('?') - 47, False),
    KeyCode(92, 92, '\\', False, ord('|') - 92, False),
    KeyCode(96, 96, '`', False, ord('~') - 96, False),
    KeyCode(45, 45, '-', False, ord('_') - 45, False),
    KeyCode(61, 61, '=', False, ord('+') - 61, False),
    KeyCode(44, 44, ',', False, ord('<') - 44, False),
    KeyCode(46, 46, '.', False, ord('>') - 46, False),
]

# Missing stuff
# PrintScreen
# ScrollLock
# Pause
# Ralt

for i in range(1, 14):
    _codes.append(KeyCode(281 + i, 111 + i, 'f{}'.format(i), True, 0, False))

for i in range(26):
    _codes.append(KeyCode(97 + i, 65 + i, chr(97 + i), False, -32, True))

_nums_shift = [')', '!', '@', '#', '$', '%', '^', '&', '*', '(']

for i in range(10):
    _codes.append(KeyCode(48 + i, 48 + i, str(i), False, (ord(_nums_shift[i]) - 48 - i), False))


_by_char = {
    character: (unix, windows, is_special, shift, capslock)
    for (unix, windows, character, is_special, shift, capslock) in _codes
}

_by_unix = {
    unix: (character, windows, is_special, shift, capslock)
    for (unix, windows, character, is_special, shift, capslock) in _codes
}

_by_windows = {
    windows: (character, unix, is_special, shift, capslock)
    for (unix, windows, character, is_special, shift, capslock) in _codes
}


def get_unix_keycode(character):
    return _by_char[character][0]


def get_windows_keycode(character):
    return _by_char[character][1]


def get_native_keycode(character):
    return get_unix_keycode(character)


def get_character_unix(unix, modifiers=None):
    if not modifiers:
        modifiers = []
    if unix not in _by_unix.keys():
        return chr(unix)
    spec = _by_unix[unix]
    character = spec[0]
    shift_applies = False
    offset = spec[3]
    if offset:
        capslock = spec[4]
        if not capslock:
            if 'shift' in modifiers:
                shift_applies = True
        else:
            if ('shift' in modifiers and not 'capslock' in modifiers)\
                    or ('capslock' in modifiers and not 'shift' in modifiers):
                shift_applies = True
    if shift_applies:
        if offset and len(character) == 1:
            character = chr(ord(character) + offset)
    return character


def get_character_windows(windows, modifiers=None):
    return _by_windows[windows][0]


def get_character_native(native, modifiers=None):
    return get_character_unix(native, modifiers)


def get_unix_from_windows(windows):
    return _by_windows[windows][1]


def get_windows_from_unix(unix):
    return _by_unix[unix][1]


def get_windows_from_native(native):
    return get_windows_from_unix(native)


def get_unix_from_native(native):
    return native


def check_is_special(native):
    if native not in _by_unix.keys():
        return False
    return _by_unix[native][2]


if __name__ == '__main__':
    for spec in _codes:
        print(spec)





from ..browser import cefpython


class JSBindingsMixin(object):
    def __init__(self):
        self._reset_js_bindings = False
        self._current_js_bindings = None
        self._js_bindings = [('test_binding', self._test_binding)]
        self._js_codes = []

    @property
    def reset_js_bindings(self):
        return self._reset_js_bindings

    @property
    def js_bindings(self):
        return self._js_bindings

    def preinstall_js_binding(self, name, target):
        self._js_bindings.append((name, target))

    def preinstall_js_code(self, code):
        self._js_codes.append(code)

    def _test_binding(self, *args, **kwargs):
        print("BINDING", args, kwargs)

    def set_js_bindings(self):
        # Needed to introduce set_js_bindings again because the freeze of sites at load took over.
        # As an example 'http://www.htmlbasix.com/popup.shtml' freezed every time. By setting the js
        # bindings again, the freeze rate is at about 35%. Check git to see how it was done, before using
        # this function ...
        # I (jegger) have to be honest, that I don't have a clue why this is acting like it does!
        # I hope simon (REN-840) can resolve this once in for all...
        #
        # ORIGINAL COMMENT:
        # When browser.Navigate() is called, some bug appears in CEF
        # that makes CefRenderProcessHandler::OnBrowserDestroyed()
        # is being called. This destroys the javascript bindings in
        # the Render process. We have to make the js bindings again,
        # after the call to Navigate() when OnLoadingStateChange()
        # is called with isLoading=False. Problem reported here:
        # http://www.magpcss.org/ceforum/viewtopic.php?f=6&t=11009
        if not self._current_js_bindings:
            self._current_js_bindings = cefpython.JavascriptBindings(bindToFrames=True, bindToPopups=True)
            for name, target in self.js_bindings:
                self._current_js_bindings.SetFunction(name, target)
        self.browser.SetJavascriptBindings(self._current_js_bindings)

    def inject_js_code(self, browser, is_loading, can_go_back, can_go_forward):
        frame = browser.browser.GetMainFrame()
        for code in self._js_codes:
            frame.ExecuteJavascript(code)




from cefkivy.handlers.base import ClientHandlerBase


class KeyboardHandler(ClientHandlerBase):
    # https://github.com/cztomczak/cefpython/blob/master/api/KeyboardHandler.md
    def OnPreKeyEvent(self, browser, event, event_handle, is_keyboard_shortcut_out):
        return False

    def OnKeyEvent(self, browser, event, event_handle):
        return False



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#===============================================================================
# Written by Rentouch 2013 - http://www.rentouch.ch
#===============================================================================

from kivy.lang import Builder
from kivy import resources
from kivy.factory import Factory
from kivy.core.window import Window

from browser import CefBrowser
from browser import cefpython
from jsdialogs import JSDialogs

import os

#Add folder to the kivy resource path list
#We are able than to just define relative paths regarding from the module directory.
resources.resource_add_path(os.path.abspath(os.path.join(os.path.dirname(__file__))))
Builder.load_file(resources.resource_find("popup.kv"))

RESOURCE_DIR = ""


class PopupBrowser(CefBrowser):

    def __init__(self, **kwargs):
        super(PopupBrowser, self).__init__(resources_dir=RESOURCE_DIR, **kwargs)

    def on_touch_down(self, touch, *kwargs):
        if not self.collide_point(*touch.pos):
            return
        if self.keyboard_mode == "global":
            self.request_keyboard()
        else:
            Window.release_all_keyboards()
            self.browser.GetFocusedFrame().ExecuteJavascript("__kivy__keyboard_requested = false;")

        touch.moving = False
        self.touches.append(touch)
        touch.grab(self)

        if len(self.touches) == 1:
            # Only mouse click when single touch
            y = self.height-touch.pos[1] + self.pos[1]
            x = touch.x - self.pos[0]
            self.browser.SendMouseClickEvent(
                x,
                y,
                cefpython.MOUSEBUTTON_LEFT,
                mouseUp=False,
                clickCount=1
            )
        return True

    def on_touch_move(self, touch, *kwargs):
        if touch.grab_current is not self:
            return

        if len(self.touches) == 1:
            # Moving
            if (abs(touch.dx) > 5 or abs(touch.dy) > 5) or touch.moving:
                y = self.height-touch.pos[1] + self.pos[1]
                x = touch.x - self.pos[0]
                self.browser.SendMouseMoveEvent(x, y, mouseLeave=False)
        else:
            # Scrolling
            touch1, touch2 = self.touches[:2]
            dx = touch2.dx / 2. + touch1.dx / 2.
            dy = touch2.dy / 2. + touch1.dy / 2.
            self.browser.SendMouseWheelEvent(touch.x, self.height-touch.pos[1], dx, -dy)
        return True

    def on_touch_up(self, touch, *kwargs):
        if touch.grab_current is not self:
            return
        y = self.height-touch.pos[1] + self.pos[1]
        x = touch.x - self.pos[0]
        self.browser.SendMouseClickEvent(
            x,
            y,
            cefpython.MOUSEBUTTON_LEFT,
            mouseUp=True, clickCount=1
        )
        self.touches.remove(touch)
        touch.ungrab(self)
        return True


class PopupController(object):

    def __init__(self, resource_dir=None):
        # Define data path
        if resource_dir:
            global RESOURCE_DIR
            RESOURCE_DIR = resource_dir

        # Create popup
        self.popup = Factory.CEFPopup()
        self.popup.bind(on_dismiss=self.on_close_popup)
        self.jsdialogs = JSDialogs()
        self.popup.browser.bind(on_js_dialog=self.jsdialogs.on_js_dialog)
        self.popup.browser.bind(on_before_unload_dialog=self.jsdialogs.on_before_unload_dialog)

    def on_before_popup(self, obj, browser, frame, targetUrl, targetFrameName,
                        popupFeatures, windowInfo, client, browserSettings, noJavascriptAccess):
        print(windowInfo, popupFeatures, targetFrameName)
        self.popup.url = targetUrl
        self.popup.open()

    def on_close_popup(self, *kwargs):
        self.popup.url = "about:blank"





import os
from kivy.logger import Logger
from kivy.properties import StringProperty

from ..browser import cefpython


class CookieManagerMixin(object):
    resources_dir = StringProperty("")

    def __init__(self, resources_dir):
        self.resources_dir = resources_dir
        # Set cookie manager
        # Determine if default resources dir should be used or a custom
        if self.resources_dir:
            resources = self.resources_dir
        else:
            resources = cefpython.GetModuleDirectory()

        Logger.debug("cefkivy: Using Resource Directory <{}>".format(resources))
        Logger.debug("cefkivy: Creating the CookieManager")
        cookie_manager = cefpython.CookieManager.GetGlobalManager()
        cookie_path = os.path.join(resources, "cookies")
        cookie_manager.SetStoragePath(cookie_path, True)

    def delete_cookie(self, url=""):
        """ Deletes the cookie with the given url. If url is empty all cookies get deleted.
        """
        cookie_manager = cefpython.CookieManager.GetGlobalManager()
        if cookie_manager:
            cookie_manager.DeleteCookies(url, "")
        else:
            print("No cookie manager found!, Can't delete cookie(s)")





from kivy.properties import StringProperty
from kivy.properties import BooleanProperty


class NavigationMixin(object):
    url = StringProperty("about:blank")
    current_url = StringProperty("")
    is_loading = BooleanProperty(True)

    def __init__(self, start_url):
        self.url = start_url
        self.register_event_type("on_loading_state_change")
        self.register_event_type("on_address_change")
        self.register_event_type("on_title_change")
        self.register_event_type("on_load_start")
        self.register_event_type("on_load_end")
        self.register_event_type("on_load_error")

    def on_url(self, instance, value):
        if self.browser and value:
            self.browser.Navigate(self.url)
            # self._reset_js_bindings = True

    def go_back(self):
        self.browser.GoBack()

    def go_forward(self):
        self.browser.GoForward()

    def on_loading_state_change(self, isLoading, canGoBack, canGoForward):
        self.is_loading = isLoading

    def on_address_change(self, frame, url):
        self.current_url = url

    def on_title_change(self, new_title):
        pass

    def on_load_start(self, frame):
        pass

    def on_load_end(self, frame, httpStatusCode):
        pass

    def on_load_error(self, frame, errorCode, errorText, failedUrl):
        pass




from cefkivy.handlers.base import ClientHandlerBase


class LoadHandler(ClientHandlerBase):
    # https://github.com/cztomczak/cefpython/blob/master/api/LoadHandler.md
    def OnLoadingStateChange(self, browser, is_loading, can_go_back, can_go_forward):
        self._widget.dispatch("on_loading_state_change", is_loading, can_go_back, can_go_forward)
        if self._widget.reset_js_bindings and not is_loading:
            self._widget.set_js_bindings()
        if is_loading and self._widget.keyboard_mode == "local":
            # Release keyboard when navigating to a new page.
            self._widget.release_keyboard()

    def OnLoadStart(self, browser, frame):
        self._widget.dispatch("on_load_start", frame)

    def OnLoadEnd(self, browser, frame, http_code):
        self._widget.dispatch("on_load_end", frame, http_code)

    def OnLoadError(self, browser, frame, error_code, error_text_out, failed_url):
        self._widget.dispatch("on_load_error", frame, error_code, error_text_out, failed_url)





from kivy.config import Config
Config.set('kivy', 'log_level', 'debug')
Config.set('kivy', 'keyboard_mode', 'systemandmulti')

from cefkivy.browser import CefBrowser, cefpython
from kivy.app import App


class CefBrowserApp(App):
    def build(self):
        return CefBrowser(start_url='http://google.com')


def run():
    CefBrowserApp().run()
    cefpython.Shutdown()


if __name__ == '__main__':
    run()




class DialogMixin(object):
    def __init__(self, dialog_target=None):
        self._dialog_target = dialog_target
        self._current_dialog = None

    @property
    def dialog_target(self):
        return self._dialog_target

    @dialog_target.setter
    def dialog_target(self, value):
        self._dialog_target = value

    def dialog_finish(self, dialog):
        self.dialog_target.remove_widget(self._current_dialog)
        self.disabled = False
        self.opacity = 1
        self._current_dialog = None

    def dialog_show(self, dialog):

        if not self.dialog_target:
            print("WARNING: Dialog Target not Specified. Auto Accepting!")
            dialog.skip()
            return

        if self._current_dialog:
            dialog.cancel()
            return

        self._current_dialog = dialog.build()
        dialog.when_done = self.dialog_finish
        self.opacity = 0.3
        self.disabled = True
        self.dialog_target.add_widget(self._current_dialog)





from ..inputs.touch import CefTouchProcessor


class TouchMixin(object):
    def __init__(self):
        self._touch_processor = CefTouchProcessor(widget=self, browser=self.browser)

    def on_touch_down(self, touch, **kwargs):
        self._touch_processor.on_touch_down(touch, **kwargs)

    def on_touch_move(self, touch, **kwargs):
        self._touch_processor.on_touch_move(touch, **kwargs)

    def on_touch_up(self, touch, **kwargs):
        self._touch_processor.on_touch_up(touch, **kwargs)





import os
from kivy.base import EventLoop
from kivy.core.window import Window
from kivy.uix.vkeyboard import VKeyboard


class FixedKeyboard(VKeyboard):
    def __init__(self, container, **kwargs):
        self._container = container
        super(FixedKeyboard, self).__init__(**kwargs)

    def setup_mode_free(self):
        """Overwrite free function to set fixed pos"""
        self.do_rotation = False
        self.do_scale = False
        self.scale = 1.2
        if not self._container:
            return
        self.width = self._container.width * 0.8
        self.center_x = self._container.center_x
        self.y = self._container.y


class KeyboardFactory(object):
    def __init__(self, container):
        self._container = container
        self._prototype = FixedKeyboard

    def __call__(self, **kwargs):
        return self._prototype(self._container, **kwargs)


class KeyboardManager(object):
    # Keyboard mode: "global" or "local".
    # 1. Global mode forwards keys to CEF all the time.
    # 2. Local mode forwards keys to CEF only when an editable
    #    control is focused (input type=text|password or textarea).
    __keyboard = None

    def __init__(self, widget, browser, keyboard_mode='local'):
        self._widget = widget
        self._browser = browser
        self._keyboard_factory = KeyboardFactory(container=self._widget)
        self._install()
        self.keyboard_mode = keyboard_mode

    def _install(self):
        self._widget.preinstall_js_binding('__kivy__keyboard_update', self.keyboard_update)
        with open(os.path.join(os.path.dirname(__file__), 'keyboard_trigger.js'), 'r') as f:
            js_code = f.read()
            self._widget.preinstall_js_code(js_code)
        Window.set_vkeyboard_class(self._keyboard_factory)

    @property
    def keyboard_mode(self):
        return self._keyboard_mode

    @keyboard_mode.setter
    def keyboard_mode(self, value):
        if value not in ('global', 'local'):
            raise ValueError("Invalid keyboard mode : {}".format(value))
        self._keyboard_mode = value
        if self._keyboard_mode == "global":
            self.request_keyboard()
        else:
            self.release_keyboard()

    def request_keyboard(self):
        if not self.__keyboard:
            print("Requesting Keyboard")
            self.__keyboard = EventLoop.window.request_keyboard(self.release_keyboard, self._widget)
            self.__keyboard.bind(on_key_down=self._widget.on_key_down)
            self.__keyboard.bind(on_key_up=self._widget.on_key_up)
            self.__keyboard.bind(on_textinput=self._widget.on_textinput)
            print("Got Keyboard ", self.__keyboard)
        self._widget.keystroke_processor.reset_all_modifiers()
        # Not sure if it is still required to send the focus
        # (some earlier bug), but it shouldn't hurt to call it.
        self._browser.SendFocusEvent(True)

    def release_keyboard(self):
        # When using local keyboard mode, do all the request
        # and releases of the keyboard through js bindings,
        # otherwise some focus problems arise.
        self._widget.keystroke_processor.reset_all_modifiers()
        if not self.__keyboard:
            return
        # If we blur the field on keyboard release, jumping between form
        # fields with tab won't work.
        # self.browser.GetFocusedFrame().ExecuteJavascript("__kivy__on_escape()")
        self.__keyboard.unbind(on_key_down=self._widget.on_key_down)
        self.__keyboard.unbind(on_key_up=self._widget.on_key_up)
        self.__keyboard.unbind(on_textinput=self._widget.on_textinput)
        print("Releasing Keyboard")
        self.__keyboard.release()
        self.__keyboard = None

    def keyboard_update(self, show, rect, attributes):
        """
        :param show: Show keyboard if true, hide if false (blur)
        :param rect: [x,y,width,height] of the input element
        :param attributes: Attributes of HTML element
        """
        if self.keyboard_mode == 'global':
            return
        if show:
            self.request_keyboard()
        else:
            self.release_keyboard()



# Import cef as first module! (it's important)

import ctypes
import sys
import os
import platform

libcef_so = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'libcef.so')
if os.path.exists(libcef_so):
    # Import local module
    ctypes.CDLL(libcef_so, ctypes.RTLD_GLOBAL)
    if 0x02070000 <= sys.hexversion < 0x03000000:
        import cefpython_py27 as cefpython
    else:
        raise Exception("Unsupported python version: %s" % sys.version)
else:
    # Import from package
    from cefpython3 import cefpython

from kivy.app import App

from kivy.graphics import Color, Rectangle
from kivy.graphics.texture import Texture
from kivy.uix.widget import Widget
from kivy.clock import Clock
from kivy.logger import Logger

from .mixins.keyboard import KeyboardMixin
from .mixins.touch import TouchMixin
from .mixins.navigation import NavigationMixin
from .mixins.cookies import CookieManagerMixin
from .mixins.security import SecurityMixin
from .mixins.dialogs import DialogMixin
from .mixins.popup import PopupMixin
from .mixins.jsbindings import JSBindingsMixin

from .handlers.display import DisplayHandler
from .handlers.download import DownloadHandler
from .handlers.jsdialog import JavascriptDialogHandler
from .handlers.keyboard import KeyboardHandler
from .handlers.lifespan import LifespanHandler
from .handlers.load import LoadHandler
from .handlers.render import RenderHandler
from .handlers.request import RequestHandler


class CefBrowser(PopupMixin,
                 DialogMixin,
                 TouchMixin,
                 KeyboardMixin,
                 JSBindingsMixin,
                 CookieManagerMixin,
                 SecurityMixin,
                 NavigationMixin,
                 Widget):
    browser = None

    # Subclass CefBrowser and override this list to implement custom
    # event handlers and behaviors. In general, when you subclass one of these,
    # call the parent handler before or after you have done whatever you need to
    # do. This will be required to correctly fire kivy events for the few event
    # types which are defined in the various Mixins. If you're developing a custom
    # embedded browser and aren't using the kivy events yourself, then you can
    # safely ignore them - most aren't used internally. Just ensure that any code
    # present in the built-in handlers is accounted for in terms of the
    # functionality it provides.
    _handlers = [
        DisplayHandler,
        DownloadHandler,
        JavascriptDialogHandler,
        KeyboardHandler,
        LifespanHandler,
        LoadHandler,
        RenderHandler,
        RequestHandler,
    ]

    def __init__(self, **kwargs):
        switches = kwargs.pop("switches", {'remote-debugging-port': '9222'})

        self.__rect = None
        self.browser = None
        start_url = kwargs.pop('start_url', 'about:blank')
        dialog_target = kwargs.pop('dialog_target', None)
        keyboard_mode = kwargs.pop('keyboard_mode', 'local')
        resources_dir = kwargs.pop("resources_dir", "")
        ssl_verification_disabled = kwargs.pop("ssl_verification_disabled", False)
        Widget.__init__(self, **kwargs)

        self.check_versions()

        # Create the base texture
        self.texture = Texture.create(size=self.size, colorfmt='rgba', bufferfmt='ubyte')
        self.texture.flip_vertical()
        with self.canvas:
            Color(1, 1, 1)
            self.__rect = Rectangle(pos=self.pos, size=self.size, texture=self.texture)

        def cef_loop(*_):
            cefpython.MessageLoopWork()
        Clock.schedule_interval(cef_loop, 0)

        settings = {
            # "debug": True,
            "log_severity": cefpython.LOGSEVERITY_WARNING,
            # "log_file": "debug.log",
            "persist_session_cookies": True,
            # "release_dcheck_enabled": True,  # Enable only when debugging.
            "locales_dir_path": os.path.join(cefpython.GetModuleDirectory(), "locales"),
            "browser_subprocess_path": "%s/%s" % (cefpython.GetModuleDirectory(), "subprocess"),
            'background_color': 0xFFFFFFFF,
        }

        Logger.debug("cefkivy: Intializing cefpython with \n"
                     "   settings : %s \n"
                     "   switches : %s", settings, switches)
        cefpython.Initialize(settings, switches)

        # Disable Windowed rendering and and bind to parent window 0
        # TODO Set a proper parent?
        # See https://github.com/cztomczak/cefpython/blob/master/api/WindowInfo.md#setasoffscreen
        windowInfo = cefpython.WindowInfo()
        windowInfo.SetAsOffscreen(0)

        SecurityMixin.__init__(self, ssl_verification_disabled)
        NavigationMixin.__init__(self, start_url)

        # Create the Synchronous Browser
        Logger.debug("cefkivy: Creating the Browser")
        self.browser = cefpython.CreateBrowserSync(windowInfo, {}, navigateUrl=self.url)

        CookieManagerMixin.__init__(self, resources_dir)

        self.browser.SendFocusEvent(True)

        Logger.debug("cefkivy: Installing Client Handlers")
        for handler in self._handlers:
            self.install_handler(handler)

        Logger.debug("cefkivy: Binding the Browser Size and Resizing")
        self.browser.WasResized()
        self.bind(size=self.realign)
        self.bind(pos=self.realign)

        JSBindingsMixin.__init__(self)
        KeyboardMixin.__init__(self, keyboard_mode=keyboard_mode)
        TouchMixin.__init__(self)
        DialogMixin.__init__(self, dialog_target)
        PopupMixin.__init__(self)

        Logger.debug("cefkivy: Setting JS Bindings")
        self.set_js_bindings()

        Logger.debug("cefkivy: Injecting JS Code")
        self.bind(on_loading_state_change=self.inject_js_code)

    def check_versions(self):
        md = cefpython.GetModuleDirectory()
        Logger.debug("cefkivy: Using cefpython from <{}>".format(md))
        ver = cefpython.GetVersion()
        Logger.info("cefkivy: CEF Python : {ver}".format(ver=ver["version"]))
        Logger.info("cefkivy:   Chromium : {ver}".format(ver=ver["chrome_version"]))
        Logger.info("cefkivy:        CEF : {ver}".format(ver=ver["cef_version"]))
        Logger.info("cefkivy:     Python : {ver} {arch}".format(
            ver=platform.python_version(),
            arch=platform.architecture()[0]))

    def install_handler(self, handler):
        Logger.debug("cefkivy: Installing ClientHandler <Class {}>".format(handler.__name__))
        self.browser.SetClientHandler(handler(self))

    def realign(self, *_):
        ts = self.texture.size
        ss = self.size
        schg = (ts[0] != ss[0] or ts[1] != ss[1])
        if schg:
            self.texture = Texture.create(size=self.size, colorfmt='rgba', bufferfmt='ubyte')
            self.texture.flip_vertical()
        if self.__rect:
            with self.canvas:
                Color(1, 1, 1)
                self.__rect.pos = self.pos
                if schg:
                    self.__rect.size = self.size
            if schg:
                self.update_rect()
        if self.browser:
            self.browser.WasResized()
            self.browser.NotifyScreenInfoChanged()
        # TODO : Update this to the new keyboard when implemented
        try:
            k = self.__keyboard.widget
            p = k.parent
            p.remove_widget(k)
            p.add_widget(k)
        except:
            pass

    def update_rect(self):
        if self.__rect:
            self.__rect.texture = self.texture





from ..components.blockdialog import PopupBlockDialog


class PopupMixin(object):
    popup = None

    def __init__(self):
        self.register_event_type("on_before_popup")
        # Logger.debug("cefkivy: Instantiating Browser Popup")
        # self.popup = CefBrowserPopup(self)

    def on_before_popup(self, browser, frame, target_url, target_frame_name, target_disposition,
                        user_gesture, popup_features, window_info_out, client, browser_settings_out,
                        no_javascript_access_out):
        # TODO Implement popups here. Suppressed for now.
        print("Opening Popup : ", target_url, user_gesture, target_disposition)
        block_dialog = PopupBlockDialog(browser=self.browser, callback=None,
                                        message_text=target_url)
        self.dialog_show(block_dialog)
        return True


# class CefBrowserPopup(Widget):
#     rx = NumericProperty(0)
#     ry = NumericProperty(0)
#     rpos = ReferenceListProperty(rx, ry)
#
#     def __init__(self, parent, *args, **kwargs):
#         super(CefBrowserPopup, self).__init__()
#         self.browser_widget = parent
#         self.__rect = None
#         self.texture = Texture.create(size=self.size, colorfmt='rgba', bufferfmt='ubyte')
#         self.texture.flip_vertical()
#         with self.canvas:
#             Color(1, 1, 1)
#             self.__rect = Rectangle(pos=self.pos, size=self.size, texture=self.texture)
#         self.bind(rpos=self.realign)
#         self.bind(size=self.realign)
#         parent.bind(pos=self.realign)
#         parent.bind(size=self.realign)
#
#     def realign(self, *args):
#         self.x = self.rx+self.browser_widget.x
#         self.y = self.browser_widget.height-self.ry-self.height+self.browser_widget.y
#         ts = self.texture.size
#         ss = self.size
#         schg = (ts[0] != ss[0] or ts[1] != ss[1])
#         if schg:
#             self.texture = Texture.create(size=self.size, colorfmt='rgba', bufferfmt='ubyte')
#             self.texture.flip_vertical()
#         if self.__rect:
#             with self.canvas:
#                 Color(1, 1, 1)
#                 self.__rect.pos = self.pos
#                 if schg:
#                     self.__rect.size = self.size
#             if schg:
#                 self.update_rect()
#
#     def update_rect(self):
#         if self.__rect:
#             self.__rect.texture = self.texture








from ..browser import cefpython
from ..components.blockdialog import BlockDialog


class CertificateErrorDialog(BlockDialog):
    def __init__(self, **kwargs):
        kwargs.setdefault('title', "SSL Certificate Error")
        kwargs['message_text'] = "The website at {} presented an invalid certificate " \
                                 "and has been blocked.\n" \
                                 "Please use the back or home buttons to return to " \
                                 "where you were.\n" \
                                 "".format(kwargs['message_text'])
        super(CertificateErrorDialog, self).__init__(**kwargs)


class SecurityMixin(object):
    def __init__(self, ssl_verification_disabled=False):
        self.ssl_verification_disabled = ssl_verification_disabled
        # Bind callback to the OnCertificateError cefpython event
        cefpython.SetGlobalClientCallback("OnCertificateError", self.on_certificate_error)

    def on_certificate_error(self, cert_error, request_url, callback):
        # Check if cert verification is disabled
        if self.ssl_verification_disabled:
            callback.Continue(True)
        else:
            block_dialog = CertificateErrorDialog(browser=self.browser, callback=None,
                                                  message_text=request_url, autoclose=0)
            self.dialog_show(block_dialog)
            callback.Continue(False)





from urllib.parse import urlparse
from .dialog import MessageDialogBase


class JSDialogBase(MessageDialogBase):
    def __init__(self, **kwargs):
        self._origin_url = kwargs.pop('origin_url')
        kwargs['title'] = "The website at {} says :".format(
            urlparse(self._origin_url).netloc
        )
        super(JSDialogBase, self).__init__(**kwargs)

    def cancel(self, *_):
        if self._when_done:
            self._when_done(self)
        self._callback.Continue(False, "")

    def ok(self, *_):
        if self._when_done:
            self._when_done(self)
        self._callback.Continue(True, "")


class JSDialogAlert(JSDialogBase):
    def __init__(self, **kwargs):
        kwargs.setdefault('button_specs', [
            ('OK', self.ok),
        ])
        super(JSDialogAlert, self).__init__(**kwargs)


class JSDialogConfirm(JSDialogBase):
    def __init__(self, **kwargs):
        kwargs.setdefault('button_specs', [
            ('OK', self.ok),
            ('Cancel', self.cancel)
        ])
        super(JSDialogConfirm, self).__init__(**kwargs)


class JSDialogPrompt(JSDialogBase):
    def __init__(self, default_prompt_text, **kwargs):
        kwargs.setdefault('button_specs', [
            ('OK', self.ok),
            ('Cancel', self.cancel)
        ])
        # TODO Add user_input support
        super(JSDialogPrompt, self).__init__(**kwargs)
        self._default_prompt_text = default_prompt_text

    def ok(self, user_input=""):
        if self._when_done:
            self._when_done(self)
        self._callback.Continue(True, user_input)




from cefkivy.handlers.base import ClientHandlerBase


class LifespanHandler(ClientHandlerBase):
    # https://github.com/cztomczak/cefpython/blob/master/api/LifespanHandler.md
    def OnBeforePopup(self, *args, **kwargs):
        result = self._widget.dispatch("on_before_popup", *args, **kwargs)
        return result



from kivy.uix.widget import Widget
from cefpython3.cefpython_py37 import PyBrowser
from ..browser import cefpython


class KivyTouchProcessor(object):
    def __init__(self, widget: Widget = None, **kwargs):
        if not widget:
            raise AttributeError("Widget is a required argument")
        self._touches = []
        self._widget = widget
        self._mouse_scroll_step = 40
        self._enable_dragging = kwargs.pop('enable_dragging', False)
        self._enable_right_click = kwargs.pop('enable_right_click', False)
        self._debug_touch = kwargs.pop('debug_touch', False)

    def _print_current_touches(self):
        print("Current Touches :")
        keys = ['button    ', 'is_dragging', 'is_scrolling', 'multitouch_sim', 'is_right_click', 'is_mouse_scrolling', 'is_double_tap']
        fstring = "  "
        ctitles = []
        for key in keys:
            if key.startswith('is_'):
                ctitle = key.replace('is_', "", 1)
            else:
                ctitle = key
            fstring += "{{:{}}}|".format(len(ctitle))
            ctitles.append(ctitle)

        print(fstring.format(*ctitles))
        for touch in self._touches:
            _spec = []
            for k in keys:
                if hasattr(touch, k.strip()):
                    _spec.append(getattr(touch, k.strip()))
                else:
                    _spec.append('-')
            print(fstring.format(*_spec))

    def _print_touch_event(self, touch, t):
        print(t, touch.button, touch.pos)
        for is_attr in []:
            try:
                print("  {:>15} : {}".format(is_attr, getattr(touch, "is_{}".format(is_attr))))
            except AttributeError:
                pass
        self._print_current_touches()
        print('----------------')

    def _translate_to_widget(self, x, y):
        # Widget coordinates === Browser coordinates
        y = self._widget.height - y + self._widget.pos[1]
        x = x - self._widget.pos[0]
        return x, y

    def _reset_touch_modifiers(self, touch):
        touch.is_dragging = False
        touch.is_scrolling = False
        touch.is_right_click = False

    def _prefilter_touch_start(self, touch):
        if self._widget.disabled:
            return True
        if not self._widget.collide_point(*touch.pos):
            return True
        if not self._enable_right_click and touch.button == 'right':
            return True

    def _prefilter_touch_continuation(self, touch):
        if touch.grab_current is not self._widget:
            return True

    def _accept_touch_start(self, touch):
        self._touches.append(touch)
        touch.grab(self._widget)

    def _touch_finish(self, touch):
        self._touches.remove(touch)
        touch.ungrab(self._widget)

    def on_touch_down(self, touch, **kwargs):
        if self._debug_touch:
            self._print_touch_event(touch, 'DOWN')
        if self._prefilter_touch_start(touch):
            return
        self._reset_touch_modifiers(touch)
        self._accept_touch_start(touch)
        return True

    def _process_touch_scroll(self, touch):
        # Scroll only if a given distance is passed once (could be right click)
        dx = sum([t.dx / len(self._touches) for t in self._touches])
        dy = sum([t.dy / len(self._touches) for t in self._touches])
        if (abs(dx) > 5 or abs(dy) > 5) or touch.is_scrolling:
            touch.is_scrolling = True
            self._dispatch_scroll_event(touch.x, self._widget.height - touch.pos[1], dx, -dy)

    def _process_touch_drag(self, touch, x, y):
        if (abs(touch.dx) > 5 or abs(touch.dy) > 5) or touch.is_dragging:
            if touch.is_dragging:
                self._dispatch_drag_move_event(x, y)
            else:
                self._dispatch_drag_start_event(x, y)
                touch.is_dragging = True

    def on_touch_move(self, touch, **kwargs):
        if self._debug_touch:
            self._print_touch_event(touch, 'MOVE')
        if self._prefilter_touch_continuation(touch):
            return
        x, y = self._translate_to_widget(touch.x, touch.y)

        if len(self._touches) == 1:
            if self._enable_dragging:
                self._process_touch_drag(touch, x, y)
            else:
                self._process_touch_scroll(touch)
        elif len(self._touches) == 2:
            self._process_touch_scroll(touch)
        return True

    def on_touch_up(self, touch, **kwargs):
        if self._debug_touch:
            self._print_touch_event(touch, 'UP')
        if self._prefilter_touch_continuation(touch):
            return
        x, y = self._translate_to_widget(touch.x, touch.y)

        if len(self._touches) == 2:
            if not touch.is_scrolling:
                self._touches[0].is_right_click = self._touches[1].is_right_click = True
                self._dispatch_right_click_event(x, y)
        else:
            if touch.is_scrolling:
                pass
            elif touch.is_dragging:
                self._dispatch_drag_end_event(x, y)
            elif touch.is_mouse_scrolling:
                if touch.button == 'scrolldown':
                    dy = +1
                elif touch.button == 'scrollup':
                    dy = -1
                else:
                    raise ValueError("Got mouse_scrolling with button {}".format(touch.button))
                self._dispatch_scroll_event(x, y, 0, dy * self._mouse_scroll_step)
            elif not touch.is_right_click:
                if touch.is_double_tap:
                    self._dispatch_double_click_event(x, y)
                else:
                    self._dispatch_single_click_event(x, y)
            elif touch.is_right_click or touch.button == 'right':
                self._dispatch_right_click_event(x, y)
        self._touch_finish(touch)
        return True

    def _dispatch_click_event(self, x, y, count):
        raise NotImplementedError

    def _dispatch_single_click_event(self, x, y):
        self._dispatch_click_event(x, y, 1)

    def _dispatch_double_click_event(self, x, y):
        self._dispatch_click_event(x, y, 2)

    def _dispatch_scroll_event(self, x, y, dx, dy):
        raise NotImplementedError

    def _dispatch_drag_start_event(self, x, y):
        raise NotImplementedError

    def _dispatch_drag_move_event(self, x, y):
        raise NotImplementedError

    def _dispatch_drag_end_event(self, x, y):
        raise NotImplementedError

    def _dispatch_right_click_event(self, x, y):
        raise NotImplementedError


class CefTouchProcessor(KivyTouchProcessor):
    def __init__(self, browser: PyBrowser = None, **kwargs):
        if not browser:
            raise AttributeError("Browser is a required argument")
        super(CefTouchProcessor, self).__init__(**kwargs)
        self._browser = browser

    def _dispatch_click_event(self, x, y, count):
        # Left Mouse Down, Left Mouse Up
        self._browser.SendMouseClickEvent(
            x, y, cefpython.MOUSEBUTTON_LEFT,
            mouseUp=False, clickCount=count
        )
        self._browser.SendMouseClickEvent(
            x, y, cefpython.MOUSEBUTTON_LEFT,
            mouseUp=True, clickCount=count
        )

    def _dispatch_scroll_event(self, x, y, dx, dy):
        self._browser.SendMouseWheelEvent(x, y, dx, dy)

    def _dispatch_drag_start_event(self, x, y):
        self._browser.SendMouseClickEvent(x, y, cefpython.MOUSEBUTTON_LEFT,
                                          mouseUp=False, clickCount=1)

    def _dispatch_drag_move_event(self, x, y):
        self._browser.SendMouseMoveEvent(x, y, mouseLeave=False)

    def _dispatch_drag_end_event(self, x, y):
        # Drag end (mouse up)
        self._browser.SendMouseClickEvent(
            x, y, cefpython.MOUSEBUTTON_LEFT,
            mouseUp=True, clickCount=1
        )

    def _dispatch_right_click_event(self, x, y):
        # Right click (mouse down, mouse up)
        self._browser.SendMouseClickEvent(
            x, y, cefpython.MOUSEBUTTON_RIGHT,
            mouseUp=False, clickCount=1
        )
        self._browser.SendMouseClickEvent(
            x, y, cefpython.MOUSEBUTTON_RIGHT,
            mouseUp=True, clickCount=1
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import setuptools
from setuptools import setup

install_reqs = [
    'cefpython3==66.0',
    'kivy',
]

import cefkivy

# setup
setup(name='cefkivy-ebs',
      version=cefkivy.__version__,
      author='Chintalagiri Shashank',
      author_email='shashank@chintal.in',
      url='https://github.com/ebs-universe/cefkivy',
      packages=setuptools.find_packages(),
      package_data={'cefkivy': ['*.kv',
                                'components/*.js']},
      python_requires='>=3.4, <3.8',
      install_requires=install_reqs,
      entry_points={
            'console_scripts': [
                  'cefkivy-example = cefkivy.example:run'
            ]},
      )






from kivy.clock import Clock
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
# TODO Make the ebs dependency installable or just include the code directly
from ebs.iot.linuxnode.widgets.colors import ColorBoxLayout
from ebs.iot.linuxnode.widgets.labels import WrappingLabel


class MessageDialogBase(object):
    def __init__(self, browser, message_text, callback,
                 button_specs=None, user_input=False, when_done=None, title=None,
                 bgcolor=None, icon=None, fgcolor=None, autoclose=0):
        self._when_done = when_done
        self._browser = browser
        self._message_text = message_text
        self._callback = callback
        self._bgcolor = bgcolor or (0.8, 0.8, 0.8, 1)
        self._fgcolor = fgcolor or (0, 0, 0, 1)
        self._icon = icon
        self._button_specs = button_specs or []
        self._user_input = user_input
        self._title = title
        self._autoclose = autoclose
        self._finished = False

    def build(self):
        if self._autoclose:
            self._message_text = self._message_text + \
                                 "\nThis message will close in {} seconds.".format(self._autoclose)

            def _autoclose(*_):
                if not self._finished:
                    self.cancel()
            Clock.schedule_once(_autoclose, self._autoclose)

        dialog_widget = ColorBoxLayout(
            bgcolor=self._bgcolor, orientation='vertical',
            size_hint=(0.5, None), pos_hint={'center_x': 0.5, 'center_y': 0.5},
            padding=[10],
        )

        if self._title:
            title = Label(text=self._title, color=self._fgcolor)
            dialog_widget.add_widget(title)

        rich_layout = BoxLayout(orientation='horizontal', size_hint_y=None)
        if self._icon:
            pass

        uix_layout = BoxLayout(orientation='vertical', spacing=10, padding=[10], size_hint_y=None)
        label = WrappingLabel(text=self._message_text, color=self._fgcolor, size_hint=(1, None))
        uix_layout.add_widget(label)

        if self._user_input:
            pass

        def _uix_resize(_, l_h):
            uix_layout.height = l_h + 20
        label.bind(height=_uix_resize)

        rich_layout.add_widget(uix_layout)

        def _rich_resize(_, u_h):
            rich_layout.height = u_h
        uix_layout.bind(height=_rich_resize)

        dialog_widget.add_widget(rich_layout)

        button_layout = BoxLayout(orientation='horizontal', spacing=10, height=35)
        for text, action in self._button_specs:
            lbutton = Button(text=text)
            lbutton.bind(on_press=action)
            button_layout.add_widget(lbutton)

        dialog_widget.add_widget(button_layout)

        def _dialog_resize(_, r_h):
            dialog_widget.height = r_h + 35 + title.height + 20
        rich_layout.bind(height=_dialog_resize)
        self._finished = False
        return dialog_widget

    def skip(self):
        self.ok()

    def cancel(self, *_):
        self._finished = True
        if self._when_done:
            self._when_done(self)
        if self._callback:
            self._callback(False)

    def ok(self, *_):
        self._finished = True
        if self._when_done:
            self._when_done(self)
        if self._callback:
            self._callback(True)

    @property
    def when_done(self):
        return self._when_done

    @when_done.setter
    def when_done(self, value):
        self._when_done = value



from cefpython3 import cefpython

from cefkivy.handlers.base import ClientHandlerBase


class RenderHandler(ClientHandlerBase):
    # https://github.com/cztomczak/cefpython/blob/master/api/RenderHandler.md
    def GetRootScreenRect(self, browser, rect_out):
        pass

    def GetViewRect(self, browser, rect_out):
        width, height = self._widget.texture.size
        rect_out.append(0)
        rect_out.append(0)
        rect_out.append(width)
        rect_out.append(height)
        return True

    def GetScreenRect(self, browser, rect_out):
        pass

    def GetScreenPoint(self, browser, view_x, view_y, screen_coordinates_out):
        pass

    def OnPopupShow(self, browser, show):
        self._widget.remove_widget(self._widget.popup)
        if show:
            self._widget.add_widget(self._widget.popup)

    def OnPopupSize(self, browser, rect_out):
        self._widget.popup.rpos = (rect_out[0], rect_out[1])
        self._widget.popup.size = (rect_out[2], rect_out[3])

    def OnPaint(self, browser, element_type, dirty_rects, paint_buffer, width, height):
        b = paint_buffer.GetString(mode="bgra", origin="top-left")

        if element_type != cefpython.PET_VIEW:
            popup = self._widget.popup
            if popup.texture.width * popup.texture.height * 4 != len(b):
                return  # prevent segfault
            popup.texture.blit_buffer(b, colorfmt='bgra', bufferfmt='ubyte')
            popup.update_rect()
            return

        if self._widget.texture.width * self._widget.texture.height * 4 != len(b):
            return  # prevent segfault
        self._widget.texture.blit_buffer(b, colorfmt='bgra', bufferfmt='ubyte')
        self._widget.update_rect()
        return

    def OnCursorChange(self, browser, cursor):
        pass

    def OnScrollOffsetChanged(self, browser):
        pass

    def OnTextSelectionChanged(self, browser, selected_text, selected_range):
        pass





from cefkivy.components.dialog import MessageDialogBase


class BlockDialog(MessageDialogBase):
    def __init__(self, **kwargs):
        kwargs.setdefault('button_specs', [
            ('OK', self.ok),
        ])
        kwargs.setdefault('autoclose', 8)
        super(BlockDialog, self).__init__(**kwargs)


class PopupBlockDialog(BlockDialog):
    def __init__(self, **kwargs):
        kwargs.setdefault('title', "Popup Blocked")
        kwargs['message_text'] = "The website attempted to open a popup to {} which was blocked." \
                                 "".format(kwargs['message_text'])
        super(PopupBlockDialog, self).__init__(**kwargs)




from cefkivy.handlers.base import ClientHandlerBase


class NavigationHandler(ClientHandlerBase):
    def OnAddressChange(self, browser, frame, url):
        self._widget.dispatch("on_address_change", frame, url)

    def OnTitleChange(self, browser, title):
        self._widget.dispatch("on_title_change", title)


class DisplayMessageHandler(ClientHandlerBase):
    def OnTooltip(self, browser, text_out):
        _ = text_out[0]
        # Let the browser deal with tooltips.
        return False

    def OnStatusMessage(self, browser, value):
        pass

    def OnConsoleMessage(self, browser, level, message, source, line):
        pass


class DisplayHandler(NavigationHandler,
                     DisplayMessageHandler):
    # https://github.com/cztomczak/cefpython/blob/master/api/DisplayHandler.md
    def __init__(self, browser_widget):
        NavigationHandler.__init__(self, browser_widget)
        DisplayMessageHandler.__init__(self, browser_widget)
